
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <unordered_set>
#include <vector>
#include "ftr.hh"
#include "makecfg.hh"

#include "u64.hh"



class CFG_Edge
{
public:
    CFG_Edge ( ulongint id , CFG_Vertex * start ,
                CFG_Vertex * end , ulongint frequency ) ;
    virtual ~CFG_Edge () {} ;
    void Export ( FILE * outfd ) ;
    void Add_Instruction ( ulongint pc , ulongint cost ) ;
    bool Merge ( CFG_Edge * other ) ;
    void Check_Instruction ( std::unordered_set<ulongint> & known ) ;

    ulongint                frequency , id , branch_cost ;
    CFG_Vertex *            start ;
    CFG_Vertex *            end ;
    std::vector<std::pair<ulongint, ulongint> > path ;
    BType                   branch_type;
} ;


class CFG_Edge_Set : public std::set<CFG_Edge *> {} ;


class CFG_Vertex
{
public:
    CFG_Vertex ( ulongint id , ulongint pc ) ;
    virtual ~CFG_Vertex () {} ;
    bool Trim () ;
    void Export ( FILE * outfd ) ;
    void Check_Frequency () ;
    void Add_Edges ( std::list<CFG_Edge *> & all_edges ) ;
    void Export_UDG ( FILE * udgf , Symbol_Table * symtab ) ;

    bool Is_Entrance () { return entrances . size () == 0 ; } ;
    bool Is_Exit () { return exits . size () == 0 ; } ;

    CFG_Edge_Set        entrances , exits ;
    ulongint            pc , id ;
    bool                udg_printed ;
} ;



Make_CFG :: Make_CFG ()
{
    entrance = 0L ;
}

Make_CFG :: ~Make_CFG ()
{
    Unmake () ;
}

void Make_CFG :: Read_Trace ( FILE * ftr0fd ,
            ulongint main_start , ulongint main_end ,
            FILE * exe_trace_out )
{
    struct FastTraceRecord      ftrp [ 1 ] ;
    struct FastTraceRecord *    ftr0 = & ftrp [ 0 ] ;
    bool                        about_to_activate = false ;
    bool                        about_to_deactivate = false ;
    bool                        active = false ;
    bool                        can_activate = true ;
    bool                        unknown_warn = false ;
    ulongint                    expect_pc = 0 ;
    ulongint                    current_pc = 0 ;
    ulongint                    source_pc = 0 ;
    ulongint                    num_repeats = 0 ;
    BType                       old_bt = BT_None ;

    bb_mark . clear () ;
    fall_through . clear () ;
    exec_frequency . clear () ;
    branch_links . clear () ;
    branch_type . clear () ;
    cost_lookup . clear () ;
    repeat_lookup . clear () ;

    while ( fread ( ftr0 , 
            sizeof ( struct FastTraceRecord ) , 1 , ftr0fd ) == 1 )
    {
        bool is_trap = ( ftr0 -> optype == 'T' ) ;
        BType bt = BT_None ;
        current_pc = ftr0 -> pc_value ;
        assert (( current_pc & 3 ) == 0 ) ;
        source_pc = expect_pc - 4 ;

        if ( is_trap )
        {
            continue ;
        }

        //fprintf ( stderr , "%"FX" is %c\n" , current_pc , ftr0 -> optype ) ;
        switch ( ftr0 -> optype )
        {
            case 'X' :      
                    bt = branch_type [ current_pc ] = BT_Serialising ;
                    break ;
            case 'B' :      
                    bt = branch_type [ current_pc ] = BT_Conditional ;
                    break ;
            case 'J' :      
                    bt = BT_Unconditional ;
                    break ;
            case '?' :
                    if ( ! unknown_warn )
                    {
                        unknown_warn = true;
                        fprintf ( stderr , 
                            "Warning: Unknown integer ALU op at 0x%"FX"\n" ,
                            current_pc ) ;
                    }
                    break ;
            default :       
                    break ;
        }
        // Keep track of costs, and the repeat counter, which is
        // very important for LDM/STM operations on ARM.
        if (( cost_lookup . count ( current_pc ) == 0 )
        || ( cost_lookup [ current_pc ] < ftr0 -> cost ))
        {
            cost_lookup [ current_pc ] = ftr0 -> cost ;
        }
        if ( repeat_lookup . count ( current_pc ) == 0 )
        {
            repeat_lookup [ current_pc ] = num_repeats ;
        } 

        // Activation mechanism - activation begins when
        // "main" calls "wcetmain", and ends when "wcetmain" returns.
        bool in_main = (( main_start <= current_pc )
            && ( current_pc < main_end )) ;
        if ( active )
        {
            if ( in_main ) 
            {
                // deactivate
                bb_mark . insert ( source_pc ) ;
                about_to_deactivate = true ;
            }
            about_to_activate = false ;
            can_activate = false ;
        } else {
            if ( in_main ) 
            {
                about_to_activate = can_activate ;
            } else if ( about_to_activate )
            {
                active = true ;
                bb_mark . insert ( current_pc ) ;
            }
        }

        if ( ! active )
        {
            continue ;
        }

        if (( exe_trace_out != 0L )
        && ( current_pc != source_pc ))
        {
            fprintf ( exe_trace_out , "%"FX"\n" , current_pc ) ;
        }

        if ( current_pc == source_pc )
        {
            // Repeat instruction (e.g. stm/ldm)
            num_repeats ++ ;
            if ( repeat_lookup [ current_pc ] < num_repeats )
            {
                repeat_lookup [ current_pc ] = num_repeats ;
            }
        } else if (( current_pc != expect_pc )
        || ( old_bt == BT_Unconditional ))
        {
            // Branch to new basic block
            num_repeats = 1 ;
            bb_mark . insert ( expect_pc ) ;
            bb_mark . insert ( current_pc ) ;

            // Record branch link
            if ( branch_links . count ( source_pc ) != 0 )
            {
                // Branch source has been seen before
                if ( branch_links [ source_pc ] . count (
                                    current_pc ) != 0 )
                {
                    // This branch has been seen before
                    branch_links [ source_pc ][ 
                                current_pc ] ++ ;
                } else {
                    // New branch destination
                    branch_links [ source_pc ][ 
                                current_pc ] = 1 ;
                }
            } else {
                // New branch source and destination
                branch_links [ source_pc ][ 
                            current_pc ] = 1 ;
            }
        } else {
            // Fall through
            num_repeats = 1 ;
            if ( old_bt != BT_None )
            {
                // There's a branch here as well
                bb_mark . insert ( current_pc ) ;
            }
            if ( fall_through . count ( source_pc ) == 0 )
            {
                fall_through [ source_pc ] = 1 ;
            } else {
                fall_through [ source_pc ] ++ ;
            }
        }

        if ( current_pc != source_pc )
        {
            if ( exec_frequency . count ( source_pc ) == 0 )
            {
                exec_frequency [ source_pc ] = 1 ;
            } else {
                exec_frequency [ source_pc ] ++ ;
            }
        }

        if ( about_to_deactivate )
        {
            active = false ;
        }
        expect_pc = current_pc + 4 ;
        old_bt = bt ;
    }
    if ( can_activate || active )
    {
        // Before main, or during main - shouldn't exit here.
        // Can't give an exact location due to buffering in the FIFO.
        fprintf ( stderr , "Warning: early exit after PC 0x%"FX"\n" ,
                current_pc ) ;
    }
    // The end
    bb_mark . erase ( 0 ) ;
}


void Make_CFG :: Make ()
{
    std::set<ulongint>::iterator            i ;
    std::map<ulongint, ulongint>::iterator  k ;
    std::map<ulongint, CFG_Vertex *>        cfg_vertex ;
    std::list<CFG_Vertex *>::iterator       l ;
    ulongint                                bb_start , bb_end , j ;
    ulongint                                uid = 1 ;

    Unmake () ;
    // Step 1 - create CFG vertexes
    for ( i = bb_mark . begin () ; i != bb_mark . end () ; i ++ )
    {
        CFG_Vertex * cv = new CFG_Vertex ( uid , (* i) ) ;
        cfg_vertex [ (* i) ] = cv ;
        all_cfg_vertexes . push_back ( cv ) ;
        uid ++ ;
    }
    // Step 2 - create CFG edges
    i = bb_mark . begin () ;
    bb_start = (* i) ;
    i ++ ;
    for ( ; i != bb_mark . end () ; i ++ )
    {
        // Get data about this basic block
        bb_end = (* i) - 4 ;
        ulongint frequency = 0 ;

        if ( exec_frequency . count ( bb_start ) == 1 )
        {
            frequency = exec_frequency [ bb_start ] ;
        }
        
        if (( frequency == 0 )
        && ( branch_links . count ( bb_end ) == 0 ))
        {
            // Never executed - dead code/gap between functions
            bb_start = (* i) ;
            continue ;
        }

        // Lookup/make vertexes
        CFG_Vertex * cv_start = cfg_vertex [ bb_start ] ;
        CFG_Vertex * cv_end = new CFG_Vertex ( uid , bb_end ) ;
        all_cfg_vertexes . push_back ( cv_end ) ;
        uid ++ ;

        // Add edge to represent basic block
        CFG_Edge * bb = new CFG_Edge ( uid , cv_start , cv_end , frequency ) ;
        uid ++ ;

        // By definition - no branches in the middle of a basic block!
        for ( j = bb_start ; j < bb_end ; j += 4 )
        {
            assert ( branch_links . count ( j ) == 0 ) ;
            assert ( exec_frequency . count ( j ) == 1 ) ;
            // Check that execution frequencies are constant within
            // each basic block
            assert ( exec_frequency [ j ] == frequency ) ;
            assert ( frequency != 0 ) ;
            bb -> Add_Instruction ( j , repeat_lookup [ j ] * cost_lookup [ j ] ) ;
        }
        bb -> Add_Instruction ( bb_end , 
                repeat_lookup [ bb_end ] * cost_lookup [ bb_end ] ) ;

        // Branch type (if any)
        if ( branch_type . count ( bb_end ) != 0 )
        {
            bb -> branch_type = branch_type [ bb_end ] ;
        } else {
            bb -> branch_type = BT_Unconditional ;
        }

        // At end of basic block - maybe fall-through to next 
        if ( fall_through . count ( bb_end ) == 1 )
        {
            ulongint bb_target = bb_end + 4 ;
            ulongint frequency = fall_through [ bb_end ] ;

            assert ( frequency != 0 ) ;
            assert ( cfg_vertex . count ( bb_target ) == 1 ) ;
            new CFG_Edge ( uid , cv_end , 
                    cfg_vertex [ bb_target ] , frequency ) ;
            uid ++ ;

            if ( bb -> branch_type == BT_Unconditional )
            {
                bb -> branch_type = BT_None ;
            }
        }
        // At end of basic block - branches handled
        if ( branch_links . count ( bb_end ) != 0 )
        {
            for ( k = branch_links [ bb_end ] . begin () ; 
                    k != branch_links [ bb_end ] . end () ; k ++ )
            {
                ulongint bb_target = (* k) . first ;
                ulongint frequency = (* k) . second ;

                assert ( cfg_vertex . count ( bb_target ) == 1 ) ;
                CFG_Edge * e = new CFG_Edge ( uid , cv_end , 
                        cfg_vertex [ bb_target ] , frequency ) ;
                e -> branch_cost = 1 ;
                uid ++ ;
            }
        }
        bb_start = (* i) ;
    }
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;

        cv -> Check_Frequency () ;
    }
    // Step 3: Trim redundant vertexes
    bool changes_made = true ;
    while ( changes_made )
    {
        changes_made = false ;
        for ( l = all_cfg_vertexes . begin () ;
                l != all_cfg_vertexes . end () ; )
        {
            CFG_Vertex * cv = (* l) ;

            if ( cv -> Trim () )
            {
                delete cv ;
                changes_made = true ;
                l = all_cfg_vertexes . erase ( l ) ;
            } else {
                l ++ ;
            }
        }
    }
    // Step 4: Check frequencies and make list of edges
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;

        cv -> Check_Frequency () ;
        cv -> Add_Edges ( all_cfg_edges ) ;
        if ( cv -> Is_Entrance () )
        {
            entrance = cv ;
        }
    }
    // Step 5: Check instructions are unique
    std::unordered_set<ulongint> known ;
    std::list<CFG_Edge *>::iterator m ;

    for ( m = all_cfg_edges . begin () ;
            m != all_cfg_edges . end () ; m ++ )
    {
        CFG_Edge * ce = (* m) ;
        ce -> Check_Instruction ( known ) ;
    }
}

void Make_CFG :: Unmake ()
{
    std::list<CFG_Edge *>::iterator i ;
    std::list<CFG_Vertex *>::iterator l ;

    for ( i = all_cfg_edges . begin () ;
            i != all_cfg_edges . end () ; i ++ )
    {
        CFG_Edge * ce = (* i) ;
        delete ce ;
    }
    for ( l = all_cfg_vertexes . begin () ;
            l != all_cfg_vertexes . end () ; l ++ )
    {
        CFG_Vertex * cv = (* l) ;
        delete cv ;
    }
    all_cfg_vertexes . clear () ;
    all_cfg_edges . clear () ;
    entrance = 0L ;
}

void Make_CFG :: Export ( FILE * outfd )
{
    std::list<CFG_Vertex *>::iterator i ;
    
    for ( i = all_cfg_vertexes . begin () ;
            i != all_cfg_vertexes . end () ; i ++ )
    {
        (* i) -> Export ( outfd ) ;
    }

    std::list<CFG_Edge *>::iterator j ;
    
    for ( j = all_cfg_edges . begin () ;
            j != all_cfg_edges . end () ; j ++ )
    {
        (* j) -> Export ( outfd ) ;
    }
}

void Make_CFG :: Export_UDG ( FILE * udgf , Symbol_Table * symtab )
{
    fputc ( '[' , udgf ) ;
    entrance -> Export_UDG ( udgf , symtab ) ;
    fprintf ( udgf , "]\n" ) ;
}


CFG_Vertex :: CFG_Vertex ( ulongint id , ulongint pc )
{
    this -> id = id ;
    this -> pc = pc ;
    udg_printed = false ;
}

bool CFG_Vertex :: Trim ()
{
    // Return true if this CFG_Vertex should be deleted
    if (( entrances . size () == 0 )
    && ( exits . size () == 0 ))
    {
        // I'm disconnected!
        return true ;
    } else if ( entrances . size () == 0 )
    {
        // I'm the program start node
        return false ;
    } else if ( exits . size () == 0 )
    {
        // I'm the program exit node
        return false ;
    } else if (( entrances . size () == 1 )
    && ( exits . size () == 1 ))
    {
        // I might be redundant.
        //  [1] ---- [2] ---- [3]   becomes [1] ---- [3]

        CFG_Vertex * cv2 = this ;
        CFG_Edge * ce12 = (* (entrances . begin ()));
        CFG_Edge * ce23 = (* (exits . begin ()));
        //CFG_Vertex * cv1 = ce12 -> start ;
        CFG_Vertex * cv3 = ce23 -> end ;

        // Sanity check
        assert ( ce12 -> end == cv2 ) ;
        assert ( ce23 -> start == cv2 ) ;

        if ( ce12 -> branch_type == BT_None )
        {
            if (( ce12 -> path . size () == 0 )
            || ( ce23 -> path . size () == 0 ))
            {
                // Merge ce12 and ce23, trim cv2 
                if ( ce12 -> Merge ( ce23 ) )
                {
                    ce12 -> end = cv3 ;
                    ce12 -> branch_type = ce23 -> branch_type ;
                    cv3 -> entrances . erase ( ce23 ) ;
                    cv3 -> entrances . insert ( ce12 ) ;
                    delete ce23 ;
                    return true ;
                }
            }
        }
    }
    return false ;
}

void CFG_Vertex :: Export ( FILE * outfd )
{
    CFG_Edge_Set::iterator i;

    fprintf ( outfd , "v %06"FX"\n" , id ) ;
    if ( entrances . size () == 0 )
    {
        fprintf ( outfd , "  entrance\n" ) ;
    }
    if ( exits . size () == 0 )
    {
        fprintf ( outfd , "  exit\n" ) ;
    }
    if ( pc != 0 )
    {
        fprintf ( outfd , "  pc %"FX"\n" , pc ) ;
    }
    for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
    {
        fprintf ( outfd , "  ie %06"FX"\n" , (* i) -> id ) ;
    }
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        fprintf ( outfd , "  oe %06"FX"\n" , (* i) -> id ) ;
    }
}

void CFG_Vertex :: Check_Frequency ()
{
    CFG_Edge_Set::iterator  i;
    ulongint                total_in = 0 ;
    ulongint                total_out = 0 ;
    ulongint                in_zero_count = 0 ;
    ulongint                out_zero_count = 0 ;
    ulongint                diff ;

    for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
    {
        total_in += (* i) -> frequency ;
        if ( (* i) -> frequency == 0 )
        {
            in_zero_count ++ ;
        }
    }
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        total_out += (* i) -> frequency ;
        if ( (* i) -> frequency == 0 )
        {
            out_zero_count ++ ;
        }
    }
    assert (( in_zero_count + out_zero_count ) <= 1 ) ;
    if ( in_zero_count == 1 )
    {
        assert ( total_out > total_in ) ;
        diff = total_out - total_in ;
        for ( i = entrances . begin () ; i != entrances . end () ; i ++ )
        {
            if ( (* i) -> frequency == 0 )
            {
                (* i) -> frequency = diff ;
                break ;
            }
        }
    } else if ( out_zero_count == 1 )
    {
        assert ( total_in > total_out ) ;
        diff = total_in - total_out ;
        for ( i = exits . begin () ; i != exits . end () ; i ++ )
        {
            if ( (* i) -> frequency == 0 )
            {
                (* i) -> frequency = diff ;
                break ;
            }
        }
    } else {
        assert (( total_in == total_out ) || Is_Entrance () || Is_Exit () ) ;
    }
}

void CFG_Vertex :: Add_Edges ( std::list<CFG_Edge *> & all_edges )
{
    CFG_Edge_Set::iterator  i;

    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        all_edges . push_back ( (* i) ) ;
    }
}

void CFG_Vertex :: Export_UDG ( FILE * udgf , Symbol_Table * symtab )
{
    if ( udg_printed )
    {
        fprintf ( udgf , "r(\"L%"FX"\")" , id ) ;
        return ;
    }
    udg_printed = true ;
    fprintf ( udgf , "l(\"L%"FX"\",n(\"C%"FU"\",[a(\"OBJECT\",\" %"FU , 
                id , id , id ) ;
    if ( pc == 0 )
    {
        //fprintf ( udgf , "" ) ;
    } else if ( symtab == 0L )
    {
        fprintf ( udgf , "%"FX"" , pc ) ;
    } else {
        symtab -> Lookup ( udgf , pc ) ;
    }
    fprintf ( udgf , "\")],[" ) ;

    CFG_Edge_Set::iterator i ;
    bool first = true ;
    
    for ( i = exits . begin () ; i != exits . end () ; i ++ )
    {
        CFG_Edge * e = (* i) ;

        if ( ! first ) 
            fprintf ( udgf , "," ) ;

        first = false ;
        fprintf ( udgf , "e(\"E%"FX"\",[a(\"OBJECT\",\"" , e -> id ) ;

        /* edge label */
        fprintf ( udgf , "%"FU" (%"FU") E%"FX , e -> frequency , 
                (ulongint) e -> path . size () , e -> id ) ;


        fprintf ( udgf , "\")]," ) ;
        e -> end -> Export_UDG ( udgf , symtab ) ;
        fprintf ( udgf , ")" ) ;
    }
    fprintf ( udgf , "]))" ) ;
}    



CFG_Edge :: CFG_Edge ( ulongint id , CFG_Vertex * start ,
                CFG_Vertex * end , ulongint frequency )
{
    this -> id = id ;
    this -> start = start ;
    this -> end = end ;
    this -> frequency = frequency ;
    this -> start -> exits . insert ( this ) ;
    this -> end -> entrances . insert ( this ) ;
    this -> branch_type = BT_None ;
    this -> branch_cost = 0 ;
}

void CFG_Edge :: Export ( FILE * outfd )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    fprintf ( outfd , "e %06"FX"\n" , id ) ;
    fprintf ( outfd , "  f %"FX"\n" , frequency ) ;
    fprintf ( outfd , "  iv %06"FX"\n" , start -> id ) ;
    fprintf ( outfd , "  ov %06"FX"\n" , end -> id ) ;
    fprintf ( outfd , "  bt " ) ;
    switch ( branch_type )
    {
        case BT_None :          fprintf ( outfd , "None" ) ; break ;
        case BT_Unconditional : fprintf ( outfd , "Unconditional" ) ; break ;
        case BT_Conditional :   fprintf ( outfd , "Conditional" ) ; break ;
        case BT_Serialising :   fprintf ( outfd , "Serialising" ) ; break ;
        default :
            assert ( 0 ) ;
            break ;
    }
    fprintf ( outfd , "\n" ) ;

    ulongint cost = branch_cost ;
    for ( i = path . begin () ; i != path . end () ; i ++ )
    {
        fprintf ( outfd , "  x %"FX"\n" , (* i) . first ) ;
        cost += (* i) . second;
    }
    fprintf ( outfd , "  cost %"FX"\n" , cost ) ;
}

void CFG_Edge :: Check_Instruction ( std::unordered_set<ulongint> & known )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    for ( i = path . begin () ; i != path . end () ; i ++ )
    {
        assert ( known . count ( (* i) . first ) == 0 ) ;
        known . insert ( (* i) . first ) ;
    }
}

void CFG_Edge :: Add_Instruction ( ulongint pc , ulongint cost )
{
    this -> path . push_back ( 
            std::pair<ulongint, ulongint> ( pc , cost ) ) ;
}

bool CFG_Edge :: Merge ( CFG_Edge * other )
{
    std::vector<std::pair<ulongint, ulongint> >::iterator i ;

    if ( other -> frequency != 0 )
    {
        if ( this -> frequency != 0 )
        {
            assert ( this -> frequency == other -> frequency ) ;
        } else {
            this -> frequency = other -> frequency ;
        }
    }
    for ( i = other -> path . begin () ; i != other -> path . end () ; i ++ )
    {
        this -> Add_Instruction ( (* i) . first , (* i) . second ) ;
    }
    return true ; // Merge successful
}



