#ifndef __CPU_VTRACE_HH__
#define __CPU_VTRACE_HH__

#include <ext/hash_map>
#include <vector>
#include <list>
#include <limits.h>
#include "base/misc.hh"

#include "config/dp_static_disambig.hh"
#include "config/dp_resync.hh"
#include "config/dp_no_forward.hh"
#include "config/dp_io_branches.hh"
#include "config/dp_allow_load_passing.hh"
#include "config/dp_ignore_cache_misses.hh"
#include "config/dp_short_tlb.hh"
#include "config/dp_paranoid.hh"
#include "config/dp_static_bpred.hh"
#include "config/dp_catch_mem_faults.hh"

typedef unsigned TraceCounter;
typedef uint64_t TracePredictionId;

class TraceFlag {
public:
    TraceFlag() : is_jump(false), is_taken(false),
        override_true_destination(false), is_end_of_trace(false),
        is_valid(false), predicted_destination(0) {}
    
    bool        is_jump;
    bool        is_taken;
    bool        override_true_destination;
    bool        is_end_of_trace;
    bool        is_valid;
    Addr        predicted_destination;
};


class MetaInstruction
{
private:
    struct TraceBranchPrediction {
        TraceFlag   flag;
        Tick        timing;
    };

    std::vector<TraceBranchPrediction>  
                    predictions;
    TraceCounter    trap;
    Addr            startPC;
    Addr            endPC;
    TraceCounter    tracePredIndex;
    bool            endReached;

public:
    MetaInstruction(Addr _start_pc = 0, unsigned _trap = 0) 
    { 
        startPC = _start_pc;
        trap = _trap;

        TraceBranchPrediction tbp;
        tbp.flag.is_valid = false;
        tbp.flag.predicted_destination = 0;
        tbp.timing = -1;
        predictions.push_back(tbp);
        endPC = 0;

        reset();
    };


public:
    unsigned getTrap () const { return trap; } ;

    // Returns true at the end of a trace.
    bool isEndOfTrace (Addr pc) const
    {
        return ( tracePredIndex >= getLength() )
                && ( pc == getEndPC() );
    };

    void setEndPC(Addr x)
    {
        endPC = x;
    };

    Addr getEndPC() const
    {
        return endPC;
    };

    // Called when a branch prediction is required.
    void nextBranchPrediction(
            bool & predict_taken,
            TraceFlag & trace_flag,
            Addr & pred_PC)
    {
        DPRINTF(Fetch, "X1 branch prediction query (%d / %d)\n", 
            (int) tracePredIndex, (int) getLength());
        assert (tracePredIndex < getLength());
        const TraceBranchPrediction & tbp = predictions[ tracePredIndex ];

        if (tbp.flag.is_end_of_trace) {
            assert ( ! endReached );
            endReached = true;
        }

        trace_flag = tbp.flag;
        predict_taken = ( trace_flag.is_jump || trace_flag.is_taken );
        pred_PC = trace_flag.predicted_destination;
        assert(trace_flag.is_valid);
        tracePredIndex ++ ;
    };

    // Called at the exit from a trace. Records the measurement, checking
    // it against existing measurements if necessary.
    void doMeasure(TraceCounter traceCounter, Tick traceTime,
            bool measureMode)
    {
        TraceBranchPrediction & tbp = predictions[ traceCounter ];
        Tick prev_measure = tbp.timing;

        if (prev_measure >= 0) {
            DPRINTF(Fetch, 
"X1 Metainst at %#x, exit %u, measured %lli earlier, %lli now.\n",
                    startPC,
                    traceCounter, 
                    prev_measure,
                    traceTime);
            
            if ( prev_measure != traceTime) {
#if DP_PARANOID
                if ( ! measureMode ) {
                    fatal("Timing changed outside of measure mode.\n");
                }
#endif
                /*if ( traceTime < tbp.timing ) */{
                    tbp.timing = traceTime;
                }
            }
        } else {
            tbp.timing = traceTime;
            DPRINTF(Fetch, "X1 Metainst at %#x, exit %u, measured %lli.\n",
                    startPC,
                    traceCounter, 
                    traceTime);
#if DP_PARANOID
            if ( ! measureMode ) {
                fatal("Execution path is unmeasured!\n");
            }
#endif
        }
    };

    void add(TraceFlag flag)
    {
        TraceBranchPrediction end = predictions.back();
        TraceBranchPrediction tbp;

        predictions.pop_back();
        tbp.flag = flag;
        tbp.timing = -1;
        predictions.push_back(tbp);
        predictions.push_back(end);
    };

    Tick getTiming(TraceCounter index) const 
    {
        return predictions[ index ].timing;
    };

    void setTiming(TraceCounter index, Tick timing) 
    {
        predictions[ index ].timing = timing;
    };

    void reset(void)
    {
        tracePredIndex = 0;
        endReached = false;
    };

    Addr getPC(void) const 
    {
        return startPC;
    };

    TraceCounter getLength(void) const
    {
        return predictions.size() - 1;
    };

    TraceFlag getType(TraceCounter index) const
    {
        return predictions[ index ].flag;
    };

    Addr getTargetPC(TraceCounter index) const
    {
        return predictions[ index ].flag.predicted_destination;
    };
};


class VirtualTrace
{
public:
    VirtualTrace() 
    { 
        TraceFlag flag;

        flag.is_jump = true;
        flag.is_end_of_trace = true;
        flag.predicted_destination = 0x120000140ULL;
        flag.is_valid = true;
        dummyMeta.add(flag);
        reset(); 
    };

private:
    class Misprediction 
    {
    public:
        Misprediction(TracePredictionId _tpid,
                    MetaInstruction * _meta,
                    TraceCounter _traceCounter,
                    Tick _startTime)
        {
            tpid = _tpid;
            meta = _meta;
            traceCounter = _traceCounter;
            startTime = _startTime;
        };

        TracePredictionId       tpid;
        MetaInstruction *       meta;
        TraceCounter            traceCounter;
        Tick                    startTime;
    
    };

    std::vector<Misprediction>  mispredictPoints;

    struct Addr_Hash
    {
        size_t operator()(Addr a) const
        { 
            return (size_t) ( a >> 2 ) ;
        };
    };


    Tick                        currentTime, startTime, measurement;
    Tick                        lastSyscall;
    Tick                        benchStartTime;
    Tick                        benchTotalTime;
    TraceCounter                traceCounter;
    __gnu_cxx::hash_map<Addr, MetaInstruction *, 
            Addr_Hash, std::equal_to<Addr> >  
                                metaInstructions;
    MetaInstruction *           currentMeta;
    MetaInstruction             dummyMeta;
    std::list<MetaInstruction *>  
                                metaInstructionsList;
    TracePredictionId           currentTracePredictionId;
    TracePredictionId           mispredictTpid;
    TraceCounter                tripPoint;
    Addr                        start;
    bool                        measureMode, trueVtrace;
    unsigned                    holdOff;

    enum VtState {
        VtReady,           
        VtFetching,
        VtEnding,
        VtRecovering,
        
        VtDisable,
        VtWaitStart
    };

    VtState                     state;

    enum BenchmarkState {
        BsInitial,
        BsWaitMispredict,
        BsTiming,
        BsDone
    };

    BenchmarkState              benchState;





public:
    // Signals
    bool isFetchPaused() const
    {
#if DP_RESYNC
        return ((state == VtEnding) || (state == VtRecovering));
#else
        return false;
#endif
    };

    // State transitions
    bool nextInstruction(Addr pc)
    {
        if ((state == VtDisable) || (state == VtWaitStart)) return false;
        if (state == VtEnding) return true;

        assert(!isFetchPaused());
        DPRINTF(Fetch, "X1 considering %#x (of %#x)\n", 
                    pc, currentMeta->getEndPC());

        if (currentMeta->isEndOfTrace(pc))
        {
            DPRINTF(Fetch, "X1 Issuing End after Breakpoint. TS=Ending.\n");
            state = VtEnding;
            return true;
        }
        return false;
    }

    bool nextBranchPrediction(
            bool & predict_taken,
            TraceFlag & trace_flag,
            Addr & pred_PC,
            TracePredictionId & tpid)
    {
        if ((state == VtDisable) || (state == VtWaitStart)) return false;

        assert(!isFetchPaused());

        tpid = currentTracePredictionId;
        currentMeta->nextBranchPrediction(predict_taken, trace_flag, pred_PC);
        mispredictPoints.push_back(Misprediction(tpid, 
                        currentMeta, traceCounter, startTime));

        bool taken = trace_flag.is_taken;

        trace_flag.override_true_destination = false;

        if (measureMode) {
            if (traceCounter == tripPoint) {
                // Change the branch or jump effects so that the
                // prediction always fails
                DPRINTF(Fetch, "X1 Trip Point reached.\n");
                if (trace_flag.is_jump) {
                    trace_flag.predicted_destination += 4;
                } else {
                    trace_flag.is_taken = ! trace_flag.is_taken;
                }
            } else {
                // Assume the prediction is correct
            }
            trace_flag.override_true_destination = true;
        }
        traceCounter++;
        currentTracePredictionId++;
        if (taken && (currentMeta->isEndOfTrace (currentMeta->getEndPC())))
        {
            DPRINTF(Fetch, "X1 Issuing End after BTK. TS=Ending.\n");
            state = VtEnding;
        }

        return true;
    };

    void notifyMispredict(TracePredictionId tpid)
    {
        if (state == VtDisable) return;
        if (tpid == 0) return;

        DPRINTF(Fetch, "X1 Misprediction %lli. TS=Recovering.\n", 
                        (long long) tpid);

        if (!trueVtrace) return;

        assert(state != VtReady);

        if (tpid < mispredictTpid) {
            mispredictTpid = tpid;
        }
        state = VtRecovering;
    }

    void notifyEmptyPipeline() 
    {
        if (holdOff != 0) {
            DPRINTF(Fetch, 
                        "X1 Pipeline empty (holdoff).\n");
        } else if (state == VtEnding) {
            DPRINTF(Fetch, 
                        "X1 Pipeline empty (Ending complete). TS=Ready.\n");
            state = VtReady;
        } else if (state == VtRecovering) {
            DPRINTF(Fetch, 
                        "X1 Pipeline empty (Recovery complete). TS=Ready.\n");
            state = VtReady;
        }
    };

    void nextFetch(Addr fetch_PC)
    {
        assert(!isFetchPaused());

        if (trueVtrace) {
            // Only allow this if we are ready 
            if (state != VtReady) {
                return ;
            }
        } else {
            // Only allow this if there is a valid vtrace
            if (metaInstructions.count(fetch_PC) != 1) {
                return;
            }
        }

        // Metainstruction change
        if (currentMeta != NULL) {
            // Record/check timing information
            recordTiming();
        }

        DPRINTF(Fetch, "X1 Issuing MI. TS=Fetching. pc=%llx\n", fetch_PC);
        if (metaInstructions.count(fetch_PC) == 1) {
            currentMeta = metaInstructions[ fetch_PC ];
            assert(currentMeta->getPC() == fetch_PC);
        } else {
            DPRINTF(Fetch, "X1 Using dummy metainstruction.\n");
            currentMeta = &dummyMeta;
            assert(trueVtrace);
        }
        currentMeta->reset();
        startTime = currentTime;
        traceCounter = 0;

        mispredictTpid = UINT_MAX;
        mispredictPoints.clear();
        state = VtFetching;
        holdOff = 3;

        if (benchState == BsWaitMispredict) {
            DPRINTF(Fetch, "X1 BS Enter Timing state.\n");
            benchState = BsTiming;
            benchStartTime = currentTime;
        }
    };

    void tick(void)
    {
        currentTime++;

        if (holdOff != 0) {
            holdOff --;
        }
        DPRINTF(Fetch, "X2 TICK %lli\n" , (long long) currentTime );
    };

    Addr getStartPC(void) const { return start; };

    bool readStartFlag(void) 
    { 
        if ((state == VtWaitStart) 
        && (currentTime > 1)) { // allow time for start up
            state = VtRecovering;
            return true;
        } else {
            return false;
        }
    };

private:
    void recordTiming(void)
    {
        Tick val = currentTime - startTime;

        if (!trueVtrace) return;

        if (currentMeta != &dummyMeta) {
            measurement = val;
        }
        if (mispredictTpid == UINT_MAX) {
            // No misprediction detected - reached the end
            DPRINTF(Fetch, "X1 Measure: no mispredict.\n");
            currentMeta->doMeasure(traceCounter, val, measureMode);
            return;
        }
        // Find misprediction
        unsigned i;
        for (i = 0; i < mispredictPoints.size(); i++) {
            const Misprediction & misp = mispredictPoints[ i ];

            if (misp.tpid == mispredictTpid) {
                DPRINTF(Fetch, "X1 Measure: mispredict at %u.\n", i);
                currentMeta->doMeasure(misp.traceCounter, val, measureMode);
                return;
            }
        }

        const char * error_text = 
                "Misprediction could not be found in current "
                "virtual instruction.\n";
#if DP_RESYNC
        fatal(error_text);
#else
        // It's just a warning - this will happen without DP_RESYNC
        DPRINTF(Fetch, error_text);
#endif
    };

public:

    // Static control functions
    void addMetaInstruction(MetaInstruction * mi)
    {
        Addr pc = mi->getPC();
        assert(metaInstructions.count(pc) == 0);
        metaInstructions[ pc ] = mi;
        metaInstructionsList.push_back(mi);
    };

    bool loadMetaInstructions(const char * fname)
    {
        const unsigned  cwsize = 4;
        FILE*           fh = fopen(fname, "rt");
        //long long       length = 0;
        long long       parameter, parameter2;
        char            line[ BUFSIZ ];
        char *          pos;
        TraceCounter    index = 0;
        MetaInstruction * meta_inst = NULL;
        TraceFlag       flag;

        memset ( & flag , 0 , sizeof ( TraceFlag ) ) ;

        if (fh == NULL) return false;

        while ( fgets(line, BUFSIZ - 1, fh) != NULL ) {
            if ( line [ 0 ] == '#' )
            {
                continue ;
            }
            pos = NULL;
            parameter = strtoll(&line[ cwsize ], &pos, 16);
            parameter2 = -1;
            if ( pos != NULL )
            {
                parameter2 = strtoll(pos, &pos, 16);
            }

            flag.is_valid = false;
            flag.is_end_of_trace = false;
            flag.is_jump = false;

            if (strncmp("STA ", line, cwsize) == 0) {
                index = 0;
                meta_inst = new MetaInstruction((Addr) parameter);
                addMetaInstruction(meta_inst);
            } else if (strncmp("LEN ", line, cwsize) == 0) {
                //length = parameter;
            } else if (strncmp("BTK ", line, cwsize) == 0) {
                flag.is_taken = true;
                flag.is_valid = true;
            } else if (strncmp("BNT ", line, cwsize) == 0) {
                flag.is_taken = false;
                flag.is_valid = true;
            } else if (strncmp("JMP ", line, cwsize) == 0) {
                flag.is_jump = true;
                flag.is_taken = true;
                flag.is_valid = true;
            } else if (strncmp("END ", line, cwsize) == 0) {
                if (parameter > 0) {
                    meta_inst->setTiming(index, parameter2);
                }
                meta_inst->setEndPC(parameter);
            } else {
                assert(0);
            }

            if (flag.is_valid) {
                flag.predicted_destination = (Addr) parameter;
                //assert(length > 0);
                assert(meta_inst != NULL);
                assert(parameter > 0);
                meta_inst->add(flag);

                if (parameter2 > 0) {
                    meta_inst->setTiming(index, parameter2);
                }

                index++;
                //length--;
            } 
        }
        return true;


    };

    void saveMetaInstructions(const char * fname)
    {
        if (!trueVtrace) return;

        FILE* fh = fopen(fname, "wt");

        if (fh == NULL) return;

        std::list<MetaInstruction *>::iterator i;

        for (i = metaInstructionsList.begin(); 
                i != metaInstructionsList.end(); i++ ) {
            MetaInstruction * meta_inst = (* i);
            fprintf(fh, "STA %llx\n", (long long) meta_inst->getPC());
            //fprintf(fh, "LEN %x\n", meta_inst->getLength());

            TraceCounter j;
            long long addr, timing;

            for (j = 0; 
                    j < meta_inst->getLength(); j++ ) {

                addr = (long long) meta_inst->getTargetPC(j);
                timing = (long long) meta_inst->getTiming(j);
                TraceFlag flag = meta_inst->getType(j);

                assert (flag.is_valid);
                if (flag.is_jump) {
                    fprintf(fh, "JMP %llx %llx\n", addr, timing);
                } else if (flag.is_taken) {
                    fprintf(fh, "BTK %llx %llx\n", addr, timing);
                } else {
                    fprintf(fh, "BNT %llx %llx\n", addr, timing);
                }
            }
            timing = (long long) meta_inst->getTiming(j);
            fprintf(fh, "END %llx %llx\n", 
                (long long) meta_inst->getEndPC(), timing);
        }
        fclose(fh);
    };


    // Dynamic control functions
    void selectStart(Addr pc, TraceCounter tp = UINT_MAX)
    {
        traceCounter = 0;
        currentTracePredictionId = 1;
        currentMeta = NULL;
        mispredictTpid = UINT_MAX;
        mispredictPoints.clear();
        tripPoint = tp;
        measurement = -1;

        start = pc;
        if (pc != 0) {
            state = VtWaitStart;
        } else {
            state = VtReady;
        }
    };

    Tick getCurrentTime() const
    {
        return currentTime;
    };

    int getMeasurement(void) const
    {
        return (int) measurement;
    };

    void reset()
    {
        selectStart(0);
        currentTime = lastSyscall = 0;
        state = VtDisable;
        measureMode = false;
        benchState = BsInitial;
        benchStartTime = 0;
        benchTotalTime = 0;
        holdOff = 0;
        trueVtrace = false;
#if DP_STATIC_DISAMBIG
#if DP_RESYNC
#if DP_NO_FORWARD
#if DP_IO_BRANCHES
#if DP_ALLOW_LOAD_PASSING
        trueVtrace = true;
#endif
#endif
#endif
#endif
#endif
    };

    bool isMeasureMode(void) const { return measureMode; };

    bool setMeasureMode(bool mm) 
    {
        measureMode = mm;
        return trueVtrace;
    };

    void syscallFlag(void) 
    {
        DPRINTF(Fetch, "X2 SYSTEM CALL\n" );
        lastSyscall = currentTime;

        if (benchState == BsInitial) {
            DPRINTF(Fetch, "X1 BS Enter Wait Mispredict state.\n");
            benchState = BsWaitMispredict;
        } else if (benchState == BsTiming) {
            benchState = BsDone;
            benchTotalTime = startTime - benchStartTime;
            DPRINTF(Fetch, "X1 BS Enter Done state: %lli - %lli cycles\n",
                (long long) startTime,
                (long long) benchStartTime);
        }
    };

    unsigned getBenchmarkTime() const
    {
        return (unsigned) benchTotalTime;
    };

    const char * getConfiguration(void) const {
        return (
#if DP_STATIC_DISAMBIG
        "DP_STATIC_DISAMBIG "
#endif
#if DP_STATIC_BPRED
        "DP_STATIC_BPRED "
#endif
#if DP_CATCH_MEM_FAULTS
        "DP_CATCH_MEM_FAULTS "
#endif
#if DP_RESYNC
        "DP_RESYNC "
#endif
#if DP_NO_FORWARD
        "DP_NO_FORWARD "
#endif
#if DP_IO_BRANCHES
        "DP_IO_BRANCHES "
#endif
#if DP_ALLOW_LOAD_PASSING
        "DP_ALLOW_LOAD_PASSING "
#endif
#if DP_IGNORE_CACHE_MISSES
        "DP_IGNORE_CACHE_MISSES "
#endif
#if DP_SHORT_TLB
        "DP_SHORT_TLB "
#endif
#if DP_PARANOID
        "DP_PARANOID "
#endif
        "." );
    };

};
    
#endif

